"""Manages the connection state to the desk."""

import asyncio
from collections.abc import Awaitable
import logging
from typing import Callable

from bleak.backends.device import BLEDevice
from bleak.exc import BleakDBusError, BleakError
from idasen import IdasenDesk

from .errors import AuthFailedError

_LOGGER = logging.getLogger(__name__)


class ConnectionManager:
    """Handles connecting to the desk. Optionally keeps retrying to connect until it succeeds."""

    def __init__(
        self,
        ble_device: BLEDevice,
        connect_callback: Callable[[], Awaitable[None]],
        disconnect_callback: Callable[[], None],
    ):
        """Init ConnectionManager."""
        self._keep_connected: bool = False
        self._connecting: bool = False
        self._retry_pending: bool = False

        self._idasen_desk: IdasenDesk = self._create_idasen_desk(ble_device)

        self._connect_callback = connect_callback
        self._disconnect_callback = disconnect_callback

    @property
    def idasen_desk(self) -> IdasenDesk:
        """The IdasenDesk instance."""
        return self._idasen_desk

    async def connect(self, retry: bool) -> None:
        """Perform the bluetooth connection to the desk."""
        self._keep_connected = True
        await self._connect(retry)

    async def disconnect(self):
        """Stop the connection manager retry task."""
        self._keep_connected = False
        if self._idasen_desk.is_connected:
            await self._idasen_desk.disconnect()

    def _create_idasen_desk(self, ble_device: BLEDevice) -> IdasenDesk:
        return IdasenDesk(
            ble_device,
            exit_on_fail=False,
            disconnected_callback=lambda bledevice: self._handle_disconnect(),
        )

    async def _connect(self, retry: bool) -> None:
        if self._connecting:
            _LOGGER.info("Connection already in progress.")
            return

        self._connecting = True
        try:
            try:
                _LOGGER.info("Connecting...")
                await self._idasen_desk.connect()
            except (TimeoutError, BleakError) as ex:
                _LOGGER.exception("Connect failed")
                if retry:
                    self._schedule_reconnect()
                    return
                raise ex

            try:
                _LOGGER.info("Pairing...")
                await self._idasen_desk.pair()
            except BleakDBusError as ex:
                _LOGGER.exception("Pair failed")
                await self._idasen_desk.disconnect()
                if retry:
                    self._schedule_reconnect()
                    return
                if ex.dbus_error == "org.bluez.Error.AuthenticationFailed":
                    raise AuthFailedError() from ex
                raise ex
            except Exception as ex:
                _LOGGER.exception("Pair failed")
                await self._idasen_desk.disconnect()
                if retry:
                    self._schedule_reconnect()
                    return
                raise ex

            await self._handle_connect()
        finally:
            self._connecting = False

    def _schedule_reconnect(self):
        RECONNECT_INTERVAL_SEC = 30
        _LOGGER.info("Will try to reconnect in %ds", RECONNECT_INTERVAL_SEC)

        if self._retry_pending:
            _LOGGER.warning("There is already a reconnect task pending")
            return
        self._retry_pending = True

        async def _reconnect():
            await asyncio.sleep(RECONNECT_INTERVAL_SEC)
            self._retry_pending = False

            if not self._keep_connected:
                _LOGGER.debug(
                    "Reconnect aborted since it should not be connected"
                    "(this could be on an older instance for an older BLEDevice)"
                )
                return

            if self.idasen_desk.is_connected:
                _LOGGER.debug("Already connected")
                return

            _LOGGER.debug("Reconnecting now")
            await self._connect(True)

        asyncio.get_event_loop().create_task(_reconnect())

    async def _handle_connect(self) -> None:
        _LOGGER.debug("Connected")
        await self._connect_callback()
        if not self._keep_connected:
            _LOGGER.info("Disconnecting since it should not be connected")
            asyncio.get_event_loop().create_task(self.disconnect())

    def _handle_disconnect(self) -> None:
        """Handle bluetooth disconnection."""
        _LOGGER.debug("Disconnected")
        if self._connecting:
            _LOGGER.debug(
                "Disconnected during connection process. No callback triggered."
            )
            return

        self._disconnect_callback()
        if self._keep_connected and not self._retry_pending:
            _LOGGER.info("Reconnecting since it should not be disconnected")
            asyncio.get_event_loop().create_task(self.connect(True))



"""Tests for idasen_ha."""

from bleak.backends.device import BLEDevice
from idasen import IdasenDesk

FAKE_BLE_DEVICE = BLEDevice("AA:BB:CC:DD:EE:FF", None, {"path": ""}, 0)


def height_percent_to_meters(percent: float):
    """Convert height from percentage to meters."""
    return IdasenDesk.MIN_HEIGHT + (IdasenDesk.MAX_HEIGHT - IdasenDesk.MIN_HEIGHT) * (
        percent / 100
    )



"""Generic test fixtures."""

from collections.abc import Awaitable
from typing import Callable, Optional
from unittest import mock
from unittest.mock import AsyncMock

from bleak import BleakClient
from idasen import IdasenDesk
import pytest


@pytest.fixture(autouse=False)
async def mock_idasen_desk():
    """Test height monitoring."""

    with mock.patch(
        "idasen_ha.connection_manager.IdasenDesk", autospec=True
    ) as patched_idasen_desk:
        patched_idasen_desk.MIN_HEIGHT = IdasenDesk.MIN_HEIGHT
        patched_idasen_desk.MAX_HEIGHT = IdasenDesk.MAX_HEIGHT

        mock_desk = patched_idasen_desk.return_value

        def mock_init(
            mac_bledevice,
            exit_on_fail: bool = False,
            disconnected_callback: Optional[Callable[[BleakClient], None]] = None,
        ):
            mock_desk.trigger_disconnected_callback = disconnected_callback
            return mock_desk

        patched_idasen_desk.side_effect = mock_init

        async def mock_connect():
            mock_desk.is_connected = True

        async def mock_disconnect():
            mock_desk.is_connected = False
            mock_desk.trigger_disconnected_callback(None)

        async def mock_monitor(callback: Callable[[float], Awaitable[None]]) -> None:
            mock_desk.trigger_monitor_callback = callback

        mock_desk.connect = AsyncMock(side_effect=mock_connect)
        mock_desk.disconnect = AsyncMock(side_effect=mock_disconnect)
        mock_desk.monitor = AsyncMock(side_effect=mock_monitor)
        mock_desk.is_connected = False
        mock_desk.is_moving = False

        yield mock_desk



"""Helper for Home Assistant Idasen Desk integration."""

from __future__ import annotations

import asyncio
import logging
from typing import Callable

from bleak.backends.device import BLEDevice
from idasen import IdasenDesk

from .connection_manager import ConnectionManager

_LOGGER = logging.getLogger(__name__)


class Desk:
    """Wrapper around the IdasenDesk from the idasen library."""

    def __init__(
        self,
        update_callback: Callable[[int | None], None] | None,
        monitor_height: bool = True,
    ) -> None:
        """Initialize the wrapper."""
        self._ble_device: BLEDevice | None = None
        self._connection_manager: ConnectionManager | None = None
        self._height: float | None = None
        self._monitor_height: bool = monitor_height

        if update_callback:
            self._update_callback = update_callback
        else:

            def empty_update_callback(height: int | None) -> None:
                pass

            self._update_callback = empty_update_callback

    async def connect(self, ble_device: BLEDevice, retry: bool = True) -> None:
        """Perform the bluetooth connection to the desk."""
        _LOGGER.debug("Connecting")

        if (
            self._connection_manager is None
            or self._ble_device is None
            or ble_device.address != self._ble_device.address
        ):
            _LOGGER.debug("Initializing idasen desk")
            await self.disconnect()
            self._ble_device = ble_device
            self._connection_manager = self._create_connection_manager(self._ble_device)

        await self._connection_manager.connect(retry)

    async def disconnect(self) -> None:
        """Disconnect from the desk."""
        if self._connection_manager:
            _LOGGER.debug("Disconnecting")
            await self._connection_manager.disconnect()

    async def move_to(self, heigh_percent: int) -> None:
        """Move the desk to a specific position."""
        _LOGGER.debug("Moving to %s", heigh_percent)
        if not self.is_connected or self._idasen_desk is None:
            _LOGGER.warning("Not connected")
            return

        if self._idasen_desk.is_moving:
            await self._idasen_desk.stop()
            # Let it settle before requesting a new move
            await asyncio.sleep(0.5)

        height = IdasenDesk.MIN_HEIGHT + (
            IdasenDesk.MAX_HEIGHT - IdasenDesk.MIN_HEIGHT
        ) * (heigh_percent / 100)

        await self._idasen_desk.move_to_target(height)

    async def move_up(self) -> None:
        """Move the desk up."""
        _LOGGER.debug("Moving up")
        await self.move_to(100)

    async def move_down(self) -> None:
        """Move the desk down."""
        _LOGGER.debug("Moving down")
        await self.move_to(0)

    async def stop(self) -> None:
        """Stop moving the desk."""
        _LOGGER.debug("Stopping")
        if not self.is_connected or self._idasen_desk is None:
            _LOGGER.warning("Not connected")
            return
        await self._idasen_desk.stop()

    async def _start_monitoring(self) -> None:
        """Start monitoring for height changes."""
        if not self.is_connected or self._idasen_desk is None:
            _LOGGER.warning("Not connected")
            return

        async def update_height(height: float) -> None:
            self._height = height
            self._update_callback(self.height_percent)

        await self._idasen_desk.monitor(update_height)

    @property
    def height(self) -> float | None:
        """The current height in meters."""
        return self._height

    @property
    def height_percent(self) -> int | None:
        """The current height in percentage."""
        if self._height is None:
            return None

        return int(
            round(
                100
                * (self._height - IdasenDesk.MIN_HEIGHT)
                / (IdasenDesk.MAX_HEIGHT - IdasenDesk.MIN_HEIGHT)
            )
        )

    @property
    def is_connected(self) -> bool:
        """True if the bluetooth connection is currently established."""
        if self._idasen_desk is None:
            return False
        # bleak `is_connected` method returns a `_DeprecatedIsConnectedReturn`,
        # so we properly cast it to bool otherwise `is_connected == True`
        # will always be False.
        return bool(self._idasen_desk.is_connected)

    @property
    def _idasen_desk(self) -> IdasenDesk | None:
        if self._connection_manager is None:
            return None
        return self._connection_manager.idasen_desk

    def _create_connection_manager(self, ble_device: BLEDevice) -> ConnectionManager:
        async def connect_callback() -> None:
            _LOGGER.debug("Connect callback called")
            if self._idasen_desk is None:
                _LOGGER.error("Desk is None after connecting")
                return

            if self._monitor_height:
                self._height = await self._idasen_desk.get_height()
                await self._start_monitoring()

            self._update_callback(self.height_percent)

        return ConnectionManager(
            ble_device,
            connect_callback=connect_callback,
            disconnect_callback=lambda: self._update_callback(self.height_percent),
        )



"""Manual testing CLI."""

# ruff: noqa: T201
import argparse
import asyncio
import logging

import aioconsole
from bleak import BleakScanner

from idasen_ha import Desk

logging.basicConfig(format="%(asctime)s [%(name)s %(levelname)s]: %(message)s")
logger = logging.getLogger("idasen_ha")
logger.setLevel(logging.DEBUG)
logging.getLogger("idasen").setLevel(logging.DEBUG)
logging.getLogger("bleak").setLevel(logging.INFO)

parser = argparse.ArgumentParser()
parser.add_argument("-a", "--address", help="Desk's bluetooth address")
args = parser.parse_args()


async def getBLEDevice(address: str):
    """Get BLE Device from address."""
    return await BleakScanner.find_device_by_address(
        address
    )  # pyright: ignore[reportGeneralTypeIssues]


def print_menu():
    """Print the menu."""
    print("\n")
    print(30 * "-", "MENU", 30 * "-")
    print("c - Connect")
    print("d - Disconnect")
    print("k - Move Up")
    print("j - Move Down")
    print("0 - Stop Moving")
    print("q - Quit")
    print("h - Print this menu")
    print(67 * "-")


async def start():
    """Start the CLI."""

    if args.address is None:
        logger.error("Desk address argument missing")
        return

    def update_callback(height: int | None):
        pass

    ble_device = await getBLEDevice(args.address)
    if ble_device is None:
        logger.error("Desk not found")
        return

    desk = Desk(update_callback)

    loop = True
    while loop:
        print_menu()
        choice = await aioconsole.ainput("Enter your choice: ")

        if choice == "h":
            print_menu()
        elif choice == "c":
            try:
                await desk.connect(ble_device, True)
            except Exception as ex:
                logger.exception(ex)
        elif choice == "d":
            await desk.disconnect()
        elif choice == "k":
            asyncio.create_task(desk.move_up())  # noqa: RUF006
        elif choice == "j":
            asyncio.create_task(desk.move_down())  # noqa: RUF006
        elif choice == "0":
            await desk.stop()
        elif choice == "q":
            loop = False
        else:
            print("Wrong option selection. Enter any key to try again..")


asyncio.run(start())



"""Tests for idasen_ha."""

import asyncio
from unittest import mock
from unittest.mock import MagicMock, Mock

from bleak.backends.device import BLEDevice
from bleak.exc import BleakDBusError, BleakError
import pytest

from idasen_ha import Desk
from idasen_ha.connection_manager import AuthFailedError

from . import FAKE_BLE_DEVICE


async def test_connect_disconnect(mock_idasen_desk: MagicMock):
    """Test connect and disconnect."""
    update_callback = Mock()
    desk = Desk(update_callback, False)

    await desk.connect(FAKE_BLE_DEVICE)
    assert desk.is_connected
    mock_idasen_desk.connect.assert_awaited()
    mock_idasen_desk.pair.assert_called()
    assert update_callback.call_count == 1

    await desk.disconnect()
    await desk.disconnect()  # double disconnect should be a no-op
    assert not desk.is_connected
    mock_idasen_desk.disconnect.assert_called()
    assert update_callback.call_count == 2


async def test_double_connect_call_with_same_bledevice(mock_idasen_desk: MagicMock):
    """Test connect being called again with the same BLEDevice, while still connecting."""
    update_callback = Mock()
    desk = Desk(update_callback, False)

    default_connect_side_effect = mock_idasen_desk.connect.side_effect

    async def connect_side_effect():
        # call the seccond `connect` while the first is ongoing
        await desk.connect(FAKE_BLE_DEVICE)
        await default_connect_side_effect()

    mock_idasen_desk.connect.side_effect = connect_side_effect

    await desk.connect(FAKE_BLE_DEVICE)
    assert desk.is_connected
    mock_idasen_desk.connect.assert_awaited()
    mock_idasen_desk.pair.assert_called()
    assert update_callback.call_count == 1


async def test_double_connect_call_with_different_bledevice():
    """Test connect being called again with a new BLEDevice, while still connecting."""

    with mock.patch(
        "idasen_ha.connection_manager.IdasenDesk", autospec=True
    ) as patched_idasen_desk:
        mock_idasen_desk = patched_idasen_desk.return_value

        async def connect_side_effect():
            # call the seccond `connect` while the first is ongoing
            mock_idasen_desk.connect.side_effect = MagicMock()
            new_ble_device = BLEDevice("AA:BB:CC:DD:EE:AA", None, None, 0)
            await desk.connect(new_ble_device)

        mock_idasen_desk.is_connected = False
        mock_idasen_desk.connect.side_effect = connect_side_effect

        update_callback = Mock()
        desk = Desk(update_callback, False)
        await desk.connect(FAKE_BLE_DEVICE)

        mock_idasen_desk.connect.assert_awaited()
        mock_idasen_desk.pair.assert_called()
        assert update_callback.call_count == 2
        assert patched_idasen_desk.call_count == 2


@mock.patch("idasen_ha.connection_manager.asyncio.sleep")
async def test_connect_called_while_retry_pending(
    sleep_mock,
    mock_idasen_desk: MagicMock,
) -> None:
    """Test connect being called while a retry is pending."""
    retry_maxed_future = asyncio.Future()
    update_callback = Mock()
    desk = Desk(update_callback, False)

    default_connect_side_effect = mock_idasen_desk.connect.side_effect

    async def sleep_side_effect(delay):
        mock_idasen_desk.connect.side_effect = default_connect_side_effect
        await desk.connect(FAKE_BLE_DEVICE)
        retry_maxed_future.set_result(None)

    sleep_mock.side_effect = sleep_side_effect

    mock_idasen_desk.connect.side_effect = TimeoutError()
    await desk.connect(FAKE_BLE_DEVICE)
    assert not desk.is_connected

    await retry_maxed_future
    assert desk.is_connected
    assert mock_idasen_desk.connect.call_count == 2
    assert update_callback.call_count == 1


async def test_connect_raises_without_auto_reconnect(mock_idasen_desk: MagicMock):
    """Test that connect raises if auto_reconnect is False."""
    desk = Desk(Mock(), False)

    mock_idasen_desk.connect.side_effect = TimeoutError()
    with pytest.raises(TimeoutError):
        await desk.connect(FAKE_BLE_DEVICE, retry=False)
    assert not desk.is_connected


@pytest.mark.parametrize(
    ("pair_exception", "raised_exception"),
    [
        (Exception(), Exception),
        (BleakDBusError("org.bluez.Error.AuthenticationFailed", ""), AuthFailedError),
    ],
)
async def test_disconnect_on_pair_failure(
    mock_idasen_desk: MagicMock, pair_exception, raised_exception
):
    """Test that disconnect is called if pair fails."""
    update_callback = Mock()
    desk = Desk(update_callback, False)

    mock_idasen_desk.pair.side_effect = pair_exception
    with pytest.raises(raised_exception):
        await desk.connect(FAKE_BLE_DEVICE, retry=False)
    assert not desk.is_connected
    mock_idasen_desk.disconnect.assert_called()
    assert update_callback.call_count == 0


@mock.patch("idasen_ha.connection_manager.asyncio.sleep")
@pytest.mark.parametrize("exception", [TimeoutError(), BleakError()])
@pytest.mark.parametrize("fail_call_name", ["connect", "pair"])
async def test_connect_exception_retry_with_disconnect(
    sleep_mock,
    mock_idasen_desk: MagicMock,
    exception: Exception,
    fail_call_name: str,
) -> None:
    """Test connect retries on connection exception."""
    TEST_RETRIES_MAX = 3
    retry_count = 0
    retry_maxed_future = asyncio.Future()

    async def sleep_handler(delay):
        nonlocal retry_count
        retry_count = retry_count + 1
        if retry_count == TEST_RETRIES_MAX:
            asyncio.get_event_loop().create_task(desk.disconnect())
            retry_maxed_future.set_result(None)

    sleep_mock.side_effect = sleep_handler

    desk = Desk(Mock(), False)

    getattr(mock_idasen_desk, fail_call_name).side_effect = exception
    await desk.connect(FAKE_BLE_DEVICE)

    await retry_maxed_future
    assert mock_idasen_desk.connect.call_count == TEST_RETRIES_MAX + 1


@mock.patch("idasen_ha.connection_manager.asyncio.sleep")
@pytest.mark.parametrize(
    "exception",
    [
        TimeoutError(),
        BleakError(),
        BleakDBusError("", []),
        BleakDBusError("org.bluez.Error.AuthenticationFailed", []),
    ],
)
@pytest.mark.parametrize("fail_call_name", ["connect", "pair"])
async def test_connect_exception_retry_success(
    sleep_mock,
    mock_idasen_desk: MagicMock,
    exception: Exception,
    fail_call_name: str,
) -> None:
    """Test connect retries on connection exception."""
    TEST_RETRIES_MAX = 3
    retry_count = 0
    retry_maxed_future = asyncio.Future()

    fail_call = getattr(mock_idasen_desk, fail_call_name)
    default_fail_call_side_effect = fail_call.side_effect

    async def sleep_handler(delay):
        nonlocal retry_count
        if retry_count == TEST_RETRIES_MAX:
            fail_call.side_effect = default_fail_call_side_effect
            retry_maxed_future.set_result(None)
        retry_count = retry_count + 1

    sleep_mock.side_effect = sleep_handler

    desk = Desk(Mock(), False)
    fail_call.side_effect = exception
    await desk.connect(FAKE_BLE_DEVICE)

    await retry_maxed_future
    assert mock_idasen_desk.connect.call_count == TEST_RETRIES_MAX + 2


async def test_reconnect_on_connection_drop(mock_idasen_desk: MagicMock):
    """Test reconnection when the connection drops."""
    update_callback = Mock()
    desk = Desk(update_callback, True)

    await desk.connect(FAKE_BLE_DEVICE)
    assert desk.is_connected
    assert update_callback.call_count == 1

    mock_idasen_desk.reset_mock()
    await mock_idasen_desk.disconnect()
    assert not desk.is_connected
    assert update_callback.call_count == 2

    await asyncio.sleep(0)
    assert desk.is_connected
    mock_idasen_desk.connect.assert_awaited()
    mock_idasen_desk.pair.assert_called()
    assert update_callback.call_count == 3



"""Tests for idasen_ha."""

from unittest import mock
from unittest.mock import MagicMock, Mock

import pytest

from idasen_ha import Desk

from . import FAKE_BLE_DEVICE, height_percent_to_meters


async def test_monitor_height(mock_idasen_desk: MagicMock):
    """Test height monitoring."""
    update_callback = Mock()
    desk = Desk(update_callback, True)

    HEIGHT_PCT_1 = 50
    HEIGHT_MTS_1 = height_percent_to_meters(HEIGHT_PCT_1)
    mock_idasen_desk.get_height.return_value = HEIGHT_MTS_1

    await desk.connect(FAKE_BLE_DEVICE)
    mock_idasen_desk.connect.assert_called()
    mock_idasen_desk.pair.assert_called()
    mock_idasen_desk.get_height.assert_called()
    update_callback.assert_called_with(HEIGHT_PCT_1)
    assert desk.height == HEIGHT_MTS_1
    assert desk.height_percent == HEIGHT_PCT_1

    HEIGHT_PCT_2 = 80
    HEIGHT_MTS_2 = height_percent_to_meters(HEIGHT_PCT_2)
    await mock_idasen_desk.trigger_monitor_callback(HEIGHT_MTS_2)
    update_callback.assert_called_with(HEIGHT_PCT_2)
    assert desk.height == HEIGHT_MTS_2
    assert desk.height_percent == HEIGHT_PCT_2


async def test_moves(mock_idasen_desk: MagicMock):
    """Test movement calls."""
    desk = Desk(None, True)
    await desk.connect(FAKE_BLE_DEVICE)

    mock_idasen_desk.is_moving = False

    HEIGHT_1 = 50
    await desk.move_to(HEIGHT_1)
    mock_idasen_desk.move_to_target.assert_called_with(
        height_percent_to_meters(HEIGHT_1)
    )

    HEIGHT_MAX = 100
    await desk.move_up()
    mock_idasen_desk.move_to_target.assert_called_with(
        height_percent_to_meters(HEIGHT_MAX)
    )

    HEIGHT_MIN = 0
    await desk.move_down()
    mock_idasen_desk.move_to_target.assert_called_with(
        height_percent_to_meters(HEIGHT_MIN)
    )


async def test_stop_before_move(mock_idasen_desk: MagicMock):
    """Test that movement is stoped before new movement starts."""
    desk = Desk(None, True)
    await desk.connect(FAKE_BLE_DEVICE)

    mock_idasen_desk.is_moving = True

    HEIGHT_1 = 50
    await desk.move_to(HEIGHT_1)
    # ensure stop() is called before move_to_target()
    mock_idasen_desk.assert_has_calls(
        [mock.call.stop(), mock.call.move_to_target(height_percent_to_meters(HEIGHT_1))]
    )


async def test_stop(mock_idasen_desk: MagicMock):
    """Test stop call."""
    desk = Desk(None, False)
    await desk.connect(FAKE_BLE_DEVICE)

    await desk.stop()
    mock_idasen_desk.stop.assert_called()


@pytest.mark.parametrize("connect_first", [True, False])
async def test_no_ops_if_not_connected(
    mock_idasen_desk: MagicMock, connect_first: bool
):
    """Test that disconnect is called if pair fails."""
    desk = Desk(Mock(), False)

    if connect_first:
        await desk.connect(FAKE_BLE_DEVICE)
        mock_idasen_desk.is_connected = False

    assert not desk.is_connected
    await desk.move_to(100)
    await desk.stop()

    mock_idasen_desk.move_to_target.assert_not_called()
    mock_idasen_desk.stop.assert_not_called()
